package njn;

import java.awt.*;

/**
 *  Defines constraints for laying out a component in a MatrixLayout
 *  container.  
 *
 *  Adding a component to a MatrixLayout container with no 
 *  MatrixConstraints object indicates to MatrixLayout that the 
 *  component should entirely fill the container.  
 *  
 *  To add a component which takes up no space at all, use
 *  "MatrixConstraints( 0, 0, 0, 0 )", or "MatrixConstraints()" (the two
 *  are equivalent).
 *
 ****/

public class MatrixConstraints extends Object
{

    public static int CELLS = 0,
                      PIXELS = 1;

    public int matrixx,
               matrixy,
               matrixwidth,
               matrixheight;
    public int pixelx,
               pixely,
               pixelwidth,
               pixelheight;

    // ------------------------------------------------------ constructors
       /**
        *  The default constructor: component takes up no space.
        ****/
    public MatrixConstraints( ) {
        matrixx = 0;
        matrixy = 0;
        matrixwidth = 0;
        matrixheight = 0;
        nullifyPixels();
    }

       /**
        *  This is probably the constructor most people will want to use.
        *  It takes the position and size of the component, in matrix cells.
        * 
        *  @param x
        *    The x-coordinate of the component's upper-left corner. 
        *    Measured in horizontal placement units.
        *
        *  @param y
        *    The y-coordinate of the component's upper-left corner. 
        *    Measured in vertical placement units.
        *  
        *  @param width
        *    The component's desired width.
        *    Measured in horizontal placement units.
        *
        *  @param height
        *    The component's desired height.
        *    Measured in vertical placement units.
        *    
        ****/
    public MatrixConstraints( int x, int y, int width, int height ) {
        matrixx = x;
        matrixy = y;
        matrixwidth = width;
        matrixheight = height;
        nullifyPixels();
    }

    /**
     *  As above, but also allows the given bounds to be interpreted 
     *  either as cells, or as pixels.
     *
     *  @param measurement_style
     *    Indicates whether the following parameters indicate cell 
     *    or pixel measurements.  Legal values are MatrixConstraints.CELLS,
     *    and MatrixConstraints.PIXELS.
     *    
     ****/
    public MatrixConstraints( int measurement_style, 
                              int x, int y, int width, int height ) {
        if ( measurement_style == CELLS ) {
            matrixx = x;
            matrixy = y;
            matrixwidth = width;
            matrixheight = height;
            nullifyPixels();
        } else {
            pixelx = x;
            pixely = y;
            pixelwidth = width;
            pixelheight = height;
            matrixx = 0;
            matrixy = 0;
            matrixwidth = 0;
            matrixheight = 0;
        }
    }


    /**
     *  As above, but uses a boolean flag to determine whether pixel
     *  measurements are to be used.
     *
     *  @param use_pixels
     *    Indicates whether the following parameters indicate pixel
     *    measurements. 
     *    
     ****/
    public MatrixConstraints( boolean use_pixels, 
                              int x, int y, int width, int height ) {
        if ( ! use_pixels ) {
            matrixx = x;
            matrixy = y;
            matrixwidth = width;
            matrixheight = height;
            nullifyPixels();
        } else {
            pixelx = x;
            pixely = y;
            pixelwidth = width;
            pixelheight = height;
            matrixx = 0;
            matrixy = 0;
            matrixwidth = 0;
            matrixheight = 0;
        }
    }

       /**
        *  Sets the component's position but not its size.
        ****/
    public MatrixConstraints( Point p ) {
        matrixx = p.x;
        matrixy = p.y;
        matrixwidth = 0;
        matrixheight = 0;
        nullifyPixels();
    }


       /**
        *  Sets the component's size but not its position.
        ****/
    public MatrixConstraints( Dimension size ) {
        matrixx = 0;
        matrixy = 0;
        matrixwidth = size.width;
        matrixheight = size.height;
        nullifyPixels();
    } // -----------------------------------------------------------------


    // ------------------------------------------------------- setPosition
    public void setPosition( int x, int y ) {
        matrixx = x;
        matrixy = y;
    } // -----------------------------------------------------------------


    // ----------------------------------------------------------- setSize
    public void setSize( int width, int height ) {
        matrixwidth = width;
        matrixheight = height;
    } // -----------------------------------------------------------------


    // -------------------------------------------------- setPixelPosition
    public void setPixelPosition( int x, int y ) {
        pixelx = x;
        pixely = y;
    } // -----------------------------------------------------------------


    // ------------------------------------------------------ setPixelSize
    public void setPixelSize( int width, int height ) {
        pixelwidth = width;
        pixelheight = height;
    } // -----------------------------------------------------------------


    // ----------------------------------------------------- nullifyPixels
    private void nullifyPixels( ) {
        pixelx = 0;
        pixely = 0;
        pixelwidth = 0;
        pixelheight = 0;
    } // -----------------------------------------------------------------

} // ===================[ End of MatrixConstraints ]======================

