package njn;

import java.awt.*;
import java.util.*;

/**
 *  First, it checks the current screen resolution and sets the default
 *  font size.
 *<p>
 *  Then, it determines the average character dimensions.
 *  These values are used in the positioning calculations.
 *
 ****/

public class MatrixLayout extends Object 
                          implements LayoutManager2
{
    Dimension screen_res;           // The current screen resolution

    int font_size;                  // The base font size to use

    int x_unit,                     // The horizontal placement unit
        y_unit;                     // The vertical placement unit
    int unit_sizes[][];

    Vector componentList,
           constraintList;

    // ------------------------------------------------------ constructors
    public MatrixLayout( ) {
        screen_res = Toolkit.getDefaultToolkit().getScreenSize();
        initFontSize();
        Font base_font = new Font("Monospaced", Font.PLAIN, font_size );
        initUnitSizes( base_font );
        initContents();
    }

    public MatrixLayout( int cell_x, int cell_y ) {
        screen_res = Toolkit.getDefaultToolkit().getScreenSize();
        x_unit = cell_x;
        y_unit = cell_y;
        initFontSize();
        initContents();
    }

    public MatrixLayout( int cell_sizes[][] ) {
        screen_res = Toolkit.getDefaultToolkit().getScreenSize();
        if ( screen_res.width <= 640 ) {
            x_unit = cell_sizes[0][0];
            y_unit = cell_sizes[0][1];
        }        
        else if ( screen_res.width <= 800 ) {
            x_unit = cell_sizes[1][0];
            y_unit = cell_sizes[1][1];
        }
        else if ( screen_res.width <= 1280 ) {
            x_unit = cell_sizes[2][0];
            y_unit = cell_sizes[2][1];
        }
        else {
            x_unit = cell_sizes[3][0];
            y_unit = cell_sizes[3][1];
        }
        initFontSize();
        initContents();
    }

    public MatrixLayout( int font_size ) {
        screen_res = Toolkit.getDefaultToolkit().getScreenSize();
        Font base_font = new Font("Monospaced", Font.PLAIN, font_size );
        initUnitSizes( base_font );
        initContents();
    } // -----------------------------------------------------------------


    private void initContents( ) {
        componentList = new Vector();
        constraintList = new Vector();
    } // -----------------------------------------------------------------


    // ------------------------------------------------------ initFontSize
       /*
        *  Set the base font size according to screen resolution:
        *
        *    640 x 480 or less       : 12
        *    800 x 600               : 14
        *    1024 x 768, 1280 x 1024 : 16
        *    Anything higher         : 18
        */
    private void initFontSize( ) {

        if ( screen_res.width <= 640 )
            font_size = 12;
        else if ( screen_res.width <= 800 )
            font_size = 14;
        else if ( screen_res.width <= 1280 )
            font_size = 16;
        else
            font_size = 18;

    } // -----------------------------------------------------------------


    // ----------------------------------------------------- initUnitSizes
       /*
        *  Define the vertical and horizontal placement units, in pixels.
        *  One unit represents half the average size (width or height) of
        *  a monospaced alphabetic character at the previously calculated 
        *  font size.  These units are therefore dependent on screen 
        *  resolution size, which is the whole point of this layout manager.
        */
    private void initUnitSizes( Font test_font ) {
        FontMetrics f = Toolkit.getDefaultToolkit().getFontMetrics( test_font );

        x_unit = (f.stringWidth("ABCDEFGHIJKLMNOPQRSTUVWXYZ") / 26) / 2;
        y_unit = f.getHeight() / 2;

    } // -----------------------------------------------------------------


    // ------------------------------------------------ addLayoutComponent
       /** 
        *  Adding a component without specifying a constraints object
        *  causes the component to fill the entire container.
        *
        *  @param name 
        *    Not used.
        *
        *  @param comp 
        *    The component to be added.
        ****/
    public void addLayoutComponent( String name, Component comp ) {
        addLayoutComponent( comp, new MatrixConstraints( 0, 0, -1, -1 ));
    }

       /**
        *  @param comp
        *    The component to be added.
        *
        *  @param constraints
        *    The MatrixConstraints object which defines the object's
        *    position and sizing behaviour.
        ****/
    public void addLayoutComponent( Component comp, Object constraints ) {
        componentList.addElement( comp );
        constraintList.addElement( constraints );
    } // -----------------------------------------------------------------


    // --------------------------------------------------- layoutContainer
       /**
        *  The magic method.
        *  The matrix cells start at 0,0 (top left cell) and end with
        *  -1,-1 (bottom right cell).
        *
        *  Each placement coordinate is multiplied by the size of the
        *  placement unit to obtain the final pixel position.
        *  In the case of negative coordinates, an "absolute" placement
        *  coordinate is calculated as follows:
        *      total number of placement units 
        *      + given placement coordinate            (a negative number)
        *      + 1                           (because we count from -1,-1)
        *  and the result is multiplied by the unit size, as usual.
        *  
        ****/
    public void layoutContainer( Container parent ) {
        Dimension container_size,     // size of container in pixels
                  matrix_size;        // size of container in matrix cells
        Component item;
        MatrixConstraints bounds;
        int x, y,                     // absolute pixel bounds of
            width, height;            //   component being shaped
        int i;

        container_size = parent.getSize();
        matrix_size = new Dimension( container_size.width / x_unit,
                                     container_size.height / y_unit );

        for ( i = 0; i < componentList.size(); i++ ) {
            item = (Component)componentList.elementAt( i );
            bounds = (MatrixConstraints)constraintList.elementAt( i );

                                    // calculate x
            if ( bounds.pixelx > 0 )
                x = bounds.pixelx;
            else if ( bounds.pixelx < 0 )
                x = container_size.width + bounds.pixelx + 1;
            else if ( bounds.matrixx < 0 )
                x = (matrix_size.width + bounds.matrixx + 1) * x_unit;
            else 
                x = bounds.matrixx * x_unit;

                                    // calculate y
            if ( bounds.pixely > 0 )
                y = bounds.pixely;
            else if ( bounds.pixely < 0 )
                y = container_size.height + bounds.pixely + 1;
            else if ( bounds.matrixy < 0 )
                y = (matrix_size.height + bounds.matrixy + 1) * y_unit;
            else 
                y = bounds.matrixy * y_unit;

                                    // calculate width
            if ( bounds.pixelwidth > 0 )
                width = bounds.pixelwidth;
            else if ( bounds.pixelwidth < 0 )
                width = container_size.width + bounds.pixelwidth + 1;
            else if ( bounds.matrixwidth < 0 )
                width = (matrix_size.width + bounds.matrixwidth + 1) * x_unit;
            else 
                width = bounds.matrixwidth * x_unit;

                                    // calculate height
            if ( bounds.pixelheight > 0 )
                height = bounds.pixelheight;
            else if ( bounds.pixelheight < 0 )
                height = container_size.height + bounds.pixelheight + 1;
            else if ( bounds.matrixheight < 0 )
                height = (matrix_size.height + bounds.matrixheight + 1) * y_unit;
            else
                height = bounds.matrixheight * y_unit;

                                    // and now shape the component
            item.setBounds( x, y, width, height );
        }

    } // -----------------------------------------------------------------


    // ----------------------------------------------- getLayoutAlignmentX
       /**
        *  What on earth is this method supposed to do?
        ****/
    public float getLayoutAlignmentX( Container target ) {
        return ( 1 );
    } // -----------------------------------------------------------------


    // ----------------------------------------------- getLayoutAlignmentY
       /**
        *  And what on earth is *this* method supposed to do?
        ****/
    public float getLayoutAlignmentY( Container target ) {
        return ( 1 );
    } // -----------------------------------------------------------------


    // -------------------------------------------------- invalidateLayout
       /**
        *  Not to mention this one...
        ****/
    public void invalidateLayout( Container target ) {
    } // -----------------------------------------------------------------


    // ------------------------------------------------- maximumLayoutSize
       /**
        *  Your guess is as good as mine...
        *  (Right now it just returns the current screen size.)
        ****/
    public Dimension maximumLayoutSize( Container target ) {
        return ( screen_res );
    } // -----------------------------------------------------------------


    // ------------------------------------------------- minimumLayoutSize
       /** 
        ****/
    public Dimension minimumLayoutSize( Container parent ) {
        int widthTotal,
            heightTotal;
        int right, 
            bottom; 
        int i;
        MatrixConstraints c;

        widthTotal = 0;
        heightTotal = 0;

        if ( ! constraintList.isEmpty() ) {
            for ( i = 0; i < constraintList.size(); i++ ) {
                c = (MatrixConstraints)constraintList.elementAt( i );
                if ( c.pixelx != 0 )
                    right = ( (c.pixelx < 0) ? 0 : c.pixelx ) +
                            ( (c.pixelwidth < 0) ? 0 : c.pixelwidth );
                else 
                    right = ( (c.matrixx < 0) ? 0 : (c.matrixx * x_unit) ) +
                            ( (c.matrixwidth < 0) ? 0 : (c.matrixwidth * x_unit ) );
                if ( c.pixely != 0 )
                    bottom = ( (c.pixely < 0) ? 0 : c.pixely ) +
                             ( (c.pixelheight < 0) ? 0 : c.pixelheight );
                else
                    bottom = ( (c.matrixy < 0) ? 0 : (c.matrixy * y_unit) ) +
                             ( (c.matrixheight < 0) ? 0 : (c.matrixheight * y_unit ) );

                if ( right > widthTotal )
                    widthTotal = right;
                if ( bottom > heightTotal )
                    heightTotal = bottom;
            }
        }
        return ( new Dimension( widthTotal, heightTotal ));

    } // -----------------------------------------------------------------


    // ----------------------------------------------- preferredLayoutSize
    public Dimension preferredLayoutSize( Container parent ) {
        return ( minimumLayoutSize( parent ));
    } // -----------------------------------------------------------------


    // --------------------------------------------- removeLayoutComponent
    public void removeLayoutComponent( Component comp ) {
        constraintList.removeElementAt( componentList.indexOf( comp ));
        componentList.removeElement( comp );
    } // -----------------------------------------------------------------


    // -------------------------------------------------------------------
       /**
        *  Return the size (in pixels) of the horizontal placement unit. 
        ****/
    public int getUnitX( ) {
        return x_unit;
    } // -----------------------------------------------------------------


    // -------------------------------------------------------------------
       /**
        *  Return the size (in pixels) of the vertical placement unit. 
        ****/
    public int getUnitY( ) {
        return y_unit;
    } // -----------------------------------------------------------------
    
    
    // -------------------------------------------------------------------
       /**
        *  Return the number of horizontal placement units corresponding
        *  to the given number of pixels.  Fractions are rounded down.
        ****/
    public int getXUnits( int pixels ) {
        return( pixels / x_unit );
    } // -----------------------------------------------------------------


    // -------------------------------------------------------------------
       /**
        *  Return the number of vertical placement units corresponding
        *  to the given number of pixels.  Fractions are rounded down.
        ****/
    public int getYUnits( int pixels ) {
        return( pixels / y_unit );
    } // -----------------------------------------------------------------


    public void setFontSize( int size ) {
        font_size = size;
    } // -----------------------------------------------------------------


    public int getFontSize() {
        return ( font_size );
    } // -----------------------------------------------------------------


} // ======================[ End of MatrixLayout ]========================

