package njn;

import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.util.*;

/**
 *  This class defines the newsreader's main application frame.
 *  
 ****/
public class NewsReader extends Frame
{
        /*
         * The following three objects will be passed to and used by most
         * other classes in the application.
         * 
         *  + "cfg" contains the global program configuration (as read
         *    from the config file);
         *    
         *  + "nntp" is an instance of the NNTP connection class which 
         *    handles all interaction with the news server.  This
         *    particular connection determines whether we are on- or off-
         *    line, reads group information, performs maintenance tasks,
         *    and (at least for now) handles all group operations for the
         *    first (and currently the only) newsgroup open.  Once 
         *    proper multithreading is implemented, additional groups
         *    will be spawning new NNTP connection objects as needed.
         *    
         *  + "news" contains the user's news and newsgroup information
         *    (as read from the newsrc file);
         *    
         */
    private Config cfg;
    private NntpConnection nntp;
    private NewsManager news;

//    ProgressTracker progress_track;

                                // package-visible interface components
    TextField status;
    MenuBar menu_bar;

    Menu programMenu,
         groupMenu,
         catchupMenu,
         refreshMenu;

    MenuItem program_quit,
             program_settings,
             program_oline,
             program_stop,
             group_open,
             group_close,
             group_subscribe,
             group_unsubscribe,
             group_post,
             catchup_all,
             catchup_current,
             refresh_all,
             refresh_current;

    Choice server_choice,
           list_view;

    Button b_connect,
               b_stop,
               b_enter,
               b_sub,
               b_refresh;

    private BorderedPanel toolbar;
    private GroupList group_list;
    private GroupWindow group_screen;
    private String message;

    // -------------------------------------------------------------------
    public NewsReader( ) {

        cfg = new Config();
        nntp = new NntpConnection();
        news = new NewsManager( cfg, nntp );

        buildInterface();
        addListeners();

//        progress_track = new ProgressTracker();

        // check to see if a news server is defined.  If not, ask for one.

        if ( cfg.start_online )
            connectToServer();
    } // -----------------------------------------------------------------


    // ---------------------------------------------------- buildInterface
    private void buildInterface() {

        this.setBackground( cfg.main_background );
        this.setFont( cfg.main_font );

        group_list = new GroupList( cfg, news, nntp );

        buildMenuBar();
        buildToolbar();
        
        status = new TextField();
        status.setEditable( false );
        status.setFont( cfg.info_font );

        this.setLayout( new BorderLayout( 6, 6 ));
        this.add("North", toolbar );
        this.add("Center", group_list );
        this.add("South", status );

        group_screen = new GroupWindow( cfg, news );
        group_screen.setVisible( false );

        group_screen.setBounds( cfg.group_window[ 0 ],
                                cfg.group_window[ 1 ],
                                cfg.group_window[ 2 ],
                                cfg.group_window[ 3 ]  );

        this.setBounds( cfg.main_window[ 0 ],
                        cfg.main_window[ 1 ],
                        cfg.main_window[ 2 ],
                        cfg.main_window[ 3 ]  );

    } // -----------------------------------------------------------------


    // ------------------------------------------------------ buildMenuBar
    private void buildMenuBar( ) {

        programMenu = new Menu("Program");
        program_settings = new MenuItem("Settings");
        program_oline = new MenuItem("Connect");
        program_stop = new MenuItem("Stop!");
        program_quit = new MenuItem("Quit");
        programMenu.add( program_settings );
        programMenu.add( program_oline );
        programMenu.add( program_stop );
        programMenu.add( program_quit );

        groupMenu = new Menu("Group");
        group_open = new MenuItem("Enter group");
        group_close = new MenuItem("Leave group");
        group_subscribe = new MenuItem("Subscribe");
        group_unsubscribe = new MenuItem("Unsubscribe");
        group_post = new MenuItem("Post article");

        catchupMenu = new Menu("Catch-up");
        catchup_current = new MenuItem("Selected group");
        catchup_all = new MenuItem("All groups");
        catchupMenu.add( catchup_current );
        catchupMenu.add( catchup_all );

        refreshMenu = new Menu("Refresh");
        refresh_current = new MenuItem("Selected group");
        refresh_all = new MenuItem("All groups");
        refreshMenu.add( refresh_current );
        refreshMenu.add( refresh_all );

        groupMenu.add( group_open );
        groupMenu.add( group_close );
        groupMenu.add( group_subscribe );
        groupMenu.add( group_unsubscribe );
        groupMenu.add( catchupMenu );
        groupMenu.add( refreshMenu );
        groupMenu.add( group_post );

        program_stop.setEnabled( false );
        group_post.setEnabled( false );
        catchupMenu.setEnabled( false );
        refreshMenu.setEnabled( false );

        menu_bar = new MenuBar();
        menu_bar.setFont( cfg.menu_font );
        menu_bar.add( programMenu );
        menu_bar.add( groupMenu );

        this.setMenuBar( menu_bar );

    } // -----------------------------------------------------------------    


    // ------------------------------------------------------ buildToolbar
    private void buildToolbar( ) {
        int i;
        FontMetrics fm = this.getFontMetrics( cfg.main_font );

        toolbar = new BorderedPanel( BorderedPanel.RAISED );
        toolbar.setLayout( new FlowLayout( FlowLayout.LEFT, 4, 3 ));

        server_choice = new Choice();
        server_choice.setBackground( cfg.text_background );
        for ( i = 0; i < cfg.news_servers.size(); i++ ) {
            server_choice.add( (String)cfg.news_servers.elementAt( i ));
        }
        server_choice.select( 0 );

        b_connect = new Button("  Connect  ");
//        b_connect.setActionCommand("Go offline");
        b_connect.setBackground( cfg.button_background );

        list_view = new Choice();
        list_view.setBackground( cfg.text_background );
        list_view.add("All Newsgroups");
        list_view.add("Subscribed Newsgroups");

        b_enter = new Button("Open");
        b_sub = new Button(" Subscribe ");
        b_refresh = new Button("Refresh");
        b_enter.setBackground( cfg.button_background );
        b_sub.setBackground( cfg.button_background );
        b_refresh.setBackground( cfg.button_background );

/*
        b_enter.setActionCommand("Enter newsgroup");
        b_sub.setActionCommand("Unsubscribe selected newsgroup");
        b_refresh.setActionCommand("Refresh all article totals");
*/

        toolbar.add( b_connect );
        toolbar.add( server_choice );
        toolbar.add( new Label(" "));
        toolbar.add( list_view );
        toolbar.add( new Label(" "));
        toolbar.add( b_enter );
        toolbar.add( b_sub );
        toolbar.add( b_refresh );

        if ( group_list.getView() == GroupList.ALL ) {
            list_view.select("All Newsgroups");
            b_sub.setLabel(" Subscribe ");
        }
        else {
            list_view.select("Subscribed Newsgroups");
            b_sub.setLabel("Unsubscribe");
        }

        group_list.requestFocus();

    } // ----------------------------------------------------------------- 

    
    // --------------------------------------------------- connectToServer
    void connectToServer( ) {
        String server_name;
        
        server_name = server_choice.getSelectedItem();
        if ( server_name == null )
            message = "No news server selected.";
        else {
            status.setText("Connecting to server \"" + server_name + "\"...");
            startWaiting(); 
            nntp.setNewsServer( server_name );
            message = nntp.connect();
            if ( nntp.connected ) {
                program_oline.setLabel("Disconnect");
//                b_connect.setActionCommand("Connected to " + server_name );
                b_connect.setLabel("Disconnect");
                server_choice.setEnabled( false );
                group_list.requestFocus();
            }
            stopWaiting();
        }

        status.setText( message );
    } // -----------------------------------------------------------------


    // ---------------------------------------------- disconnectFromServer
    void disconnectFromServer( ) {
        status.setText("Closing connection...");
        nntp.disconnect();
        if ( ! nntp.connected ) {
            program_oline.setLabel("Connect");
//            b_connect.setActionCommand("Connect to selected news server");
            b_connect.setLabel(" Connect ");
            b_connect.validate();
            message = "Disconnected successfully";
            server_choice.setEnabled( true );
        }
        status.setText( message );
    } // -----------------------------------------------------------------


    // --------------------------------------------------------- openGroup
    void openGroup( String listEntry ) {
        StringTokenizer gname;
        String groupName;
        NewsGroup group;

        if ( listEntry != null ) {
            program_oline.setEnabled( false );
            group_open.setEnabled( false );
            catchup_all.setEnabled( false );
            group_close.setEnabled( true );
            gname = new StringTokenizer( listEntry );

            startWaiting();
            try {
                gname.nextToken();
                groupName = gname.nextToken();
                status.setText("Opening group \"" + groupName +
                               "\".  Please wait...");
                group = new NewsGroup( groupName, nntp );

//                progress_track.setGoal( group.number_articles );
//                progress_track.start();

                group_screen.openNewsGroup( group );
                group_screen.show();

//                progress_track.stop();
            } catch( NoSuchElementException NSEe ) {}
            stopWaiting();
        }
    } // -----------------------------------------------------------------


    // -------------------------------------------------------- closeGroup
    void closeGroup( ) {
        program_oline.setEnabled( true );
        group_open.setEnabled( false );
        group_close.setEnabled( true );
        catchup_all.setEnabled( true );
        status.setText("Group list");
        group_screen.setVisible( false );
    } // -----------------------------------------------------------------


    // ------------------------------------------------------ addListeners
    private void addListeners( ) {
        
        this.addWindowListener(
            new WindowAdapter( ) {
                public void windowClosing( WindowEvent e ) {
                    quitProgram();
                }
            }
        );

        b_connect.addActionListener(
            new ActionListener( ) {
                public void actionPerformed( ActionEvent e ) {
                    if ( nntp.connected )
                        disconnectFromServer();
                    else connectToServer();
                }
            }
        );

        group_open.addActionListener(
            new ActionListener( ) {
                public void actionPerformed( ActionEvent e ) {
                    openGroup( group_list.selectedGroup() );
                }
            }
        );

        group_close.addActionListener(
            new ActionListener( ) {
                public void actionPerformed( ActionEvent e ) {
                    closeGroup();
                }
            }
        );

        refresh_all.addActionListener(
            new ActionListener( ) {
                public void actionPerformed( ActionEvent e ) {
                    group_list.getGroups();
                }
            }
        );

        program_quit.addActionListener(
            new ActionListener( ) {
                public void actionPerformed( ActionEvent e ) {
                    quitProgram();
                }
            }
        );

        program_oline.addActionListener(
            new ActionListener( ) {
                public void actionPerformed( ActionEvent e ) {
                    if ( nntp.connected )
                        disconnectFromServer();
                    else connectToServer();
                }
            }
        );

        b_enter.addActionListener(
            new ActionListener( ) {
                public void actionPerformed( ActionEvent e ) {
                    openGroup( group_list.selectedGroup() );
                }
            }
        );

        list_view.addItemListener(
            new ItemListener() {
                public void itemStateChanged( ItemEvent e ) {
                    if ( list_view.getSelectedItem().equals("All Newsgroups")) {
                        group_list.setView( GroupList.ALL );
                        b_sub.setLabel(" Subscribe ");
                    }
                    else {
                        group_list.setView( GroupList.SUBSCRIBED );
                        b_sub.setLabel("Unsubscribe");
                    }
		    startWaiting();
		    status.setText("Reading group list...");
                    group_list.getGroups();
		    status.setText("");
		    stopWaiting();
                }
            }
        );

        b_refresh.addActionListener(
            new ActionListener() {
                public void actionPerformed( ActionEvent e ) {
                    if ( nntp.connected ) {
                        if ( group_list.getView() == GroupList.SUBSCRIBED ) {
                            startWaiting();
                            group_list.refreshList( news.getSubscribedGroups() );
                            stopWaiting();
                        }
                    }
                }
            }
        );

        b_sub.addActionListener(
            new ActionListener() {
                public void actionPerformed( ActionEvent e ) {
                    if ( group_list.getView() == GroupList.ALL )
                        news.classifyGroup( group_list.selectedGroup(), true );
                    else
                        news.classifyGroup( group_list.selectedGroup(), false );
                }
            }
        );

                                  // --------------- Events from GroupList
        group_list.addActionListener(
            new ActionListener( ) {
                public void actionPerformed( ActionEvent e ) {
                    openGroup( group_list.selectedGroup() );
                }
            }
        );
                                // --------------- Events from GroupWindow
        group_screen.addWindowListener(
            new WindowAdapter( ) {
                public void windowClosing( WindowEvent e ) {
                    group_screen.closeNewsGroup();
                    group_screen.setVisible( false );
                    closeGroup();
                }
            }
        );

        group_screen.group_close.addActionListener(
            new ActionListener( ) {
                public void actionPerformed( ActionEvent e ) {
                    group_screen.closeNewsGroup();
                    group_screen.setVisible( false );
                    closeGroup();
                }
            }
        );

    } // -----------------------------------------------------------------


    // ------------------------------------------------------ startWaiting  
      /**
       *  Should pass a string for status as a parameter.
       *
       ****/
    void startWaiting( ) {
        this.setCursor( Cursor.getPredefinedCursor( Cursor.WAIT_CURSOR ));
        toolbar.setCursor( Cursor.getPredefinedCursor( Cursor.WAIT_CURSOR ));
        group_list.setCursor( Cursor.getPredefinedCursor( Cursor.WAIT_CURSOR ));
        status.setCursor( Cursor.getPredefinedCursor( Cursor.WAIT_CURSOR ));
    } // -----------------------------------------------------------------


    // ------------------------------------------------------- stopWaiting  
      /**
       *  Should clear status, too.
       *
       ****/
    void stopWaiting( ) {
        this.setCursor( Cursor.getDefaultCursor() );
        toolbar.setCursor( Cursor.getDefaultCursor() );
        group_list.setCursor( Cursor.getDefaultCursor() );
        status.setCursor( Cursor.getDefaultCursor() );
    } // -----------------------------------------------------------------


    // ------------------------------------------------------- quitProgram
       /*
        * Method which allows us to exit gracefully (closing the
        * connections etc.).
        * 
        */
    void quitProgram( ) {
        if ( nntp.connected ) {
            disconnectFromServer();
        }
                // Really should only do this if config was changed...
        status.setText("Saving configuration...");
        try {
            cfg.saveConfiguration();
        } catch( IOException save_error ) {
            System.out.println("Could not write configuration file:\n  " 
                               + save_error.getMessage()                 );
        }
        dispose();
        System.exit( 0 );
    } // -----------------------------------------------------------------


    
    // =================================================== ProgressMonitor
       /**
        *  (Member class)
        *
        ****
    class ProgressTracker extends Thread
    {
        int expected_total = 0;


        // ------------------------------------------------------- setGoal
        void setGoal( int goal ) {
            expected_total = goal;
        } // -------------------------------------------------------------


        // ---------------------------------------------- run (overridden)
        public void run( ) {
            while ( true ) {
                if ( nntp.socket_progress > 0 )
                    status.setText("Read " 
                                   + Integer.toString( nntp.socket_progress )
                                   + " of " 
                                   + Integer.toString( expected_total )
                                   + " articles.");
                try {
                    ProgressTracker.this.sleep( 250 );
                } catch( InterruptedException Ie ) {;}
            }
        } // -------------------------------------------------------------

    } // =================[ End of ProgressMonitor ]======================
*/
    
} // =======================[ End of NewsReader ]=========================

