package njn;

import java.io.*;
import java.net.*;
import java.util.*;

/**
 *  A simple protocol handler for the NNTP (NetNews) protocol.
 *  Should eventually manage multiple connections (sockets).
 *  
 ****/
public class NntpConnection extends Object
{
    public static final int TCP_PORT = 119;
    
                            // Response codes defined by the NNTP protocol
    public static final int 
                                /** 
                                 *  Help text (a list of valid commands)
                                 *  follows, in multi-line format. 
                                 *  Returned in response to the HELP 
                                 *  command.
                                 ****/
        NNTP_HELP_TEXT = 100,

                                /** 
                                 *  Connection to news server has been 
                                 *  established; posting is allowed on 
                                 *  this server.
                                 ****/
        NNTP_CONNECTED_POSTING_OK = 200,

                                /** 
                                 *  Connection to news server has been
                                 *  established; this server does not
                                 *  allow messages to be posted.
                                 ****/
        NNTP_CONNECTED_NO_POSTING = 201,

                                /** 
                                 *  The server has acknowledged that the
                                 *  client is a slave (multi-user) server,
                                 *  rather than a single user.  Returned
                                 *  in response to the SLAVE command.
                                 *  Server-server mode only.
                                 ****/
        NNTP_SLAVE_OK = 202,

                                /** 
                                 *  NNTP connection closed; session has
                                 *  terminated normally.  Returned in
                                 *  response to the QUIT command.
                                 ****/
        NNTP_CONNECTION_CLOSED = 205,

                                /**
                                 *  The requested newsgroup has been
                                 *  selected and queried.  Article
                                 *  summary data follows on the same line.
                                 *  Returned in response to the GROUP
                                 *  command.
                                 ****/
        NNTP_NEWSGROUP_SELECTED = 211,

                                /**
                                 *  The list of all newsgroups available
                                 *  follows, as multi-line text.  Returned
                                 *  in response to the LIST command.
                                 ****/
        NNTP_NEWSGROUP_LIST = 215,

                                /**
                                 *  Requested article found.  Article
                                 *  number and message-ID follow on the
                                 *  same line.  Article header and body
                                 *  follow as multi-line text.  Returned
                                 *  in response to the ARTICLE command.
                                 ****/
        NNTP_ARTICLE_FOLLOWS = 220,

                                /**
                                 *  Requested article found.  Article
                                 *  number and message-ID follow on the
                                 *  same line.  Article header (only)
                                 *  follows as multi-line text.  Returned
                                 *  in response to the BODY command.
                                 ****/
        NNTP_HEADER_FOLLOWS = 221,

                                /**
                                 *  Requested article found.  Article
                                 *  number and message-ID follow on the
                                 *  same line.  Article body (only)
                                 *  follows as multi-line text.  Returned
                                 *  in response to the BODY command.
                                 ****/
        NNTP_BODY_FOLLOWS = 222,

                                /**
                                 *  Requested article found.  Article
                                 *  number and Message-ID follow on the
                                 *  same line.  Article text (header and 
                                 *  body) must be requested seperately
                                 *  if desired.  
                                 ****/
        NNTP_ARTICLE_FOUND = 223,

                                /**
                                 *  The XOVER command is acknowledged;
                                 *  group overview data follows as text.
                                 ****/
        NNTP_XOVER_DATA_FOLLOWS = 224,

                                /**
                                 *  Message-IDs of all new articles 
                                 *  (since the specified timestamp) in the 
                                 *  requested group follow, as multi-line
                                 *  text.  Returned in response to the
                                 *  NEWNEWS command.
                                 ****/
        NNTP_NEW_ARTICLES = 230,

                                /**
                                 *  List of all new newsgroups created on
                                 *  the server since the specified 
                                 *  timestamp follows, as multi-line text.
                                 *  Returned in response to the NEWGROUPS
                                 *  command.
                                 ****/
        NNTP_NEW_NEWSGROUPS = 231,

                                /**
                                 *  Article (to be propagated) was
                                 *  successfully transferred from client
                                 *  to server.  Server-server mode only.
                                 ****/
        NNTP_ARTICLE_TRANSFERRED = 235,

                                /**
                                 *  Article from client was successfully
                                 *  posted.
                                 ****/
        NNTP_ARTICLE_POSTED = 240,

                                /**
                                 *  Slave server is requested to transmit
                                 *  the specified article, which is to be 
                                 *  propagated.  Returned in response to
                                 *  the IHAVE command.  Server-server mode
                                 *  only.
                                 ****/
        NNTP_TRANSFER_ARTICLE = 335,

                                /**
                                 *  Posting client should now transmit the
                                 *  article to be posted.  Returned in
                                 *  response to the POST command.
                                 ****/
        NNTP_POST_ARTICLE = 340,

                                /**
                                 *  The news server has discontinued 
                                 *  service (for some reason, such as
                                 *  by instruction of the operator).
                                 ****/
        NNTP_SERVICE_DISCONTINUED = 400,

                                /**
                                 *  The client has requested a newsgroup
                                 *  which does not exist on the server.
                                 ****/
        NNTP_NO_SUCH_NEWSGROUP = 411,

                                /**
                                 *  No newsgroup is currently selected.
                                 *  A group must be selected before the 
                                 *  requsted operation may be carried out.
                                 ****/
        NNTP_NO_NEWSGROUP_SELECTED = 412,

                                /**
                                 *  There is no article within the requested
                                 *  context.  Either the current-article 
                                 *  pointer is not set; or an empty article 
                                 *  range was requested.
                                 ****/
        NNTP_NO_ARTICLE_SELECTED = 420,

                                /**
                                 *  The current newsgroup has no articles 
                                 *  after the current one.
                                 ****/
        NNTP_NO_NEXT_ARTICLE = 421,

                                /**
                                 *  The current newsgroup has no articles 
                                 *  before the current one.
                                 ****/
        NNTP_NO_PREVIOUS_ARTICLE = 422,

                                /**
                                 *  There is no article in the current
                                 *  group corresponding to the requested
                                 *  article number.
                                 ****/
        NNTP_NO_SUCH_NUMBER = 423,

                                /**
                                 *  The requested article (requested by
                                 *  message-ID?) was not found.
                                 ****/
        NNTP_NO_SUCH_ARTICLE = 430,

                                /**
                                 *  The server does not need the specified
                                 *  article.  Returned in response to the
                                 *  IHAVE command.  Server-server mode only.
                                 ****/
        NNTP_ARTICLE_NOT_WANTED = 435,

                                /**
                                 *  An attempt to transfer an article to
                                 *  the server failed.  Another attempt
                                 *  should be made later.  Server-server
                                 *  mode only.
                                 ****/
        NNTP_TRANSFER_FAILED = 436,

                                /**
                                 *  The transferred article was rejected.
                                 *  No further attempts should be made.
                                 *  Server-server mode only.
                                 ****/
        NNTP_ARTICLE_REJECTED = 437,

                                /**
                                 *  The client attempted to post an 
                                 *  article when posting is forbidden.
                                 *  Returned by the POST command.
                                 ****/
        NNTP_POSTING_FORBIDDEN = 440,

                                /**
                                 *  An attempt to post an article to the
                                 *  server failed.  Returned by the POST
                                 *  command.
                                 ****/
        NNTP_POSTING_FAILED = 441,

                                /** 
                                 *  An invalid command was sent.
                                 ****/
        NNTP_COMMAND_NOT_RECOGNIZED = 500,

                                /**
                                 *  A command had improper syntax.
                                 ****/
        NNTP_COMMAND_SYNTAX_ERROR = 501,

                                /**
                                 *  Permission was denied or some
                                 *  access restriction was encountered.
                                 ****/
        NNTP_PERMISSION_DENIED = 502,

                                /**
                                 *  A program fault occurred, and the
                                 *  requested command was not performed.
                                 ****/
        NNTP_PROGRAM_FAULT = 503;

                                /**
                                 *  This on a line by itself (that is,
                                 *  with <cr><lf> on either side) denotes
                                 *  the end of multi-line text coming from
                                 *  the server.
                                 ****/
    public static final String NNTP_END_OF_TEXT = ".";

                                /**
                                 *  This boolean flag indicates whether
                                 *  or not we are currently connected
                                 *  to an NNTP server.
                                 ****/
    public boolean connected;
    public boolean stop_signalled;
    
    public int socket_progress;

    NntpSocket nntp_socket;
    String nntp_server;
    String server_response;

    String temp_file;
    PrintWriter temp_out;

    // -------------------------------------------------------------------
    public NntpConnection( ) {
        nntp_socket = new NntpSocket();
        nntp_server = "none";
        connected = false;
        stop_signalled = false;
        socket_progress = 0;
    }

    public NntpConnection( String news_server ) {
        nntp_socket = new NntpSocket();
        nntp_server = news_server;
        connected = false;
        stop_signalled = false;
        socket_progress = 0;
    } // -----------------------------------------------------------------


    // ----------------------------------------------------- setNewsServer
    public void setNewsServer( String news_server ) {
        nntp_server = news_server;
        socket_progress = 0;
    } // -----------------------------------------------------------------


    // -------------------------------------------------------------------
    public String connect( ) {
        if ( nntp_server.equals("none"))
            server_response = "News server has not been set.";
        else {
            try {
                server_response = nntp_socket.openConnection( nntp_server, TCP_PORT );
                if ( checkResponseCode( server_response, NNTP_CONNECTED_POSTING_OK ) ||
                    checkResponseCode( server_response, NNTP_CONNECTED_NO_POSTING )    )
                { 
                    connected = true; 
                }
            } catch( IOException any ) {
                    /*
                     * Just catch any IO exception - could be unknown host, broken
                     * socket, whatever - and return an error.  Most probable cause
                     * is not being network connected.
                     *
                     */
                server_response = "Could not establish connection.";
            } catch( NullPointerException runtime ) {
                            /*
                             * Same deal as above.  NullPointerException
                             * is also a possibility.
                             *
                             */
                server_response = "Could not establish connection.";
            }
        }
        return ( server_response );
    } // -----------------------------------------------------------------


    // -------------------------------------------------------------------
    public void disconnect( ) {
        server_response = nntp_socket.closeConnection();
        if ( checkResponseCode( server_response, NNTP_CONNECTION_CLOSED ))
            connected = false;
    } // -----------------------------------------------------------------


    // -------------------------------------------------------------------
       /**
        *  Not yet implemented.
        *  
        ****/
    public String askForHelp( ) {
        return ("Help");
    } // -----------------------------------------------------------------


    // -------------------------------------------------------------------
    public String getGroup( String groupname ) {
        server_response = nntp_socket.transmit("GROUP " + groupname );
        return ( server_response );
    } // -----------------------------------------------------------------


    // ----------------------------------------------------------- getNews
       /**
        *  Update the news database for the current group.  (Identify
        *  new articles and download the associated article headers.)
        *  
        *  @param group The name of the newsgroup being updated.
        *
        *  @return A vector containing the headers retrieved.
        *
        ****/
    public Vector getNews( String group ) {
        String group_stats,
               err_code,
               article_range;
        Vector article_list;

        group_stats = getGroup( group );
        StringTokenizer g = new StringTokenizer( group_stats, " ");
        err_code = g.nextToken();
        if ( checkResponseCode( err_code, NNTP_NEWSGROUP_SELECTED )) {
            g.nextToken();
            article_range = g.nextToken() + "-";
            article_list = getArticleList( group, group_stats, article_range );
        }
        else {
            article_list = new Vector();
            article_list.addElement( group_stats );
        }
        return ( article_list );
    }


    public Vector getNews( String group, String article_range ) {
        String group_stats,
               err_code;
        Vector article_list;

        group_stats = getGroup( group );
        if ( checkResponseCode( group_stats, NNTP_NEWSGROUP_SELECTED )) {
            article_list = getArticleList( group, group_stats, article_range );
        }
        else {
            article_list = new Vector();
            article_list.addElement( group_stats );
        }
        return ( article_list );
    } // -----------------------------------------------------------------


    // -------------------------------------------------------- getArticle
    public String getArticle( String identifier ) {
        server_response = sendCommand("ARTICLE " + identifier );
        return ( server_response );
    } // -----------------------------------------------------------------


    // ---------------------------------------------------- getArticleList
    private Vector getArticleList( String group, 
                                   String overview, 
                                   String range     ) 
    {
        Vector list;

                                                // try XOVER first
        list = downloadXOVER( group, range );
                                                // then XHDR
                                                // if all else fails...
//        list = downloadHeaders( group, overview );
        return ( list );
    } // -----------------------------------------------------------------

    
    // ----------------------------------------------------- downloadXOVER
       /*
        *  Build the article list by querying the newsgroup's overview
        *  database, using the XOVER extension to the NNTP protocol.
        *  By far the fastest and most efficient method.  Alas, not all
        *  NNTP servers support XOVER.
        *  
        */
    private Vector downloadXOVER( String group, String range ) {

        // Syntax of XOVER is:
        //    XOVER [range]
        // where "range" is one of:
        //    <article number>
        //    <article number>-
        //    <article number>-<article number>
        // and defaults to the current article number if not specified.
        //
        // The response format is:
        //    224 whatever
        //    <article number>\t<subject>\t<author>\t<date>\t<message-id>\t<references>\t<byte count>\t<line count>[\t<optional headers>]
        //    .
        // or
        //    412 no newsgroup specified
        // or
        //    420 no articles in specified range
        // or
        //    502 client does not have permission to download
        // Should also implement "LIST OVERVIEW.FMT".

        Vector article_list;
        FileReader overview_file;
        BufferedReader overview;
        String next_entry;

        article_list = new Vector();
        server_response = sendCommand("XOVER " + range );
        if ( checkResponseCode( server_response, NNTP_XOVER_DATA_FOLLOWS )) {
            try {
                overview_file = new FileReader( temp_file );
                overview = new BufferedReader( overview_file );
                do {
                    next_entry = getLine( overview );
                    if ( next_entry != null )
                        article_list.addElement( next_entry );
                } while ( next_entry != null );
                try {
                    overview_file.close();
                } catch( IOException IOe ) {;}
            } catch( FileNotFoundException FNFe ) {;}
        }
        else 
            article_list.addElement("FAILED");

        return ( article_list );
    } // -----------------------------------------------------------------


    // ------------------------------------------------------ downloadXHDR
       /*
        *  Included mostly for completeness, since AFAIK most modern NNTP 
        *  servers implement XOVER, and I suspect those that don't aren't
        *  likely to implement XHDR either.  All the same, it's worth 
        *  implementing if it means avoiding the long & painful method...
        *  
        */
    private Vector downloadXHDR( String group, String g_data ) {
        return ( new Vector() );
    } // -----------------------------------------------------------------


    // --------------------------------------------------- downloadHeaders
       /*
        *  The "long, hard method".  Builds the article overview list by 
        *  systematically finding, downloading, and parsing all the 
        *  article headers in the group, one by one.
        *  
        *  This takes a long time (and was a nuisance to code!), but
        *  fortunately is rarely necessary, assuming that the NNTP server 
        *  on the other end implements either of the XOVER or XHDR 
        *  extensions.
        *  
        */
    private Vector downloadHeaders( String group, String g_data ) {
        StringTokenizer overview = new StringTokenizer( g_data );
        StringTokenizer rtok;
        String err_code,
               first_article,
               current,         // current article number
               est_count;
        int latest, first;
        int count = 0,
            total = 0;
        Vector article_list;

        article_list = new Vector();
        try {
            err_code = overview.nextToken();
            if ( ! checkResponseCode( err_code, NNTP_NO_SUCH_NEWSGROUP )) {
                est_count = overview.nextToken();
                first_article = overview.nextToken();
                try {
                    total = Integer.parseInt( est_count );
                    first = Integer.parseInt( first_article );
                } catch( NumberFormatException NFe ) {}
                server_response = sendCommand("NEXT");
                while ( server_response.startsWith( Integer.toString( NNTP_ARTICLE_FOUND ))) {
                    count++;
                    rtok = new StringTokenizer( server_response );
                    rtok.nextToken();
                    current = rtok.nextToken();
                    sendCommand("HEAD " + current );
                    article_list.addElement( parseHeader( temp_file, current ));
                    server_response = sendCommand("NEXT");
                }
            }
        } catch( NoSuchElementException NSEe ) {}
        return ( article_list );
    } // -----------------------------------------------------------------


    // ------------------------------------------------------- parseHeader
       /*
        * Open file with name headerFile for character reading.
        * We have to retrieve the following info from each header:
        *   "Subject: %s"
        *   "From: %s"
        *   "Date: %s"
        *   "X-References: %s"
        *   "Message-ID: %s"
        *   "Lines: %s"
        * 
        */
    private String parseHeader( String headerFileName, String number ) {
        FileReader headerFile;
        BufferedReader header;
        StringTokenizer htok;
        String headerLine,
               headerField,
               msg_id,
               msg_anum,
               msg_refs,
               msg_bytes,
               msg_lines,
               msg_date,
               msg_author,
               msg_subject,
               msg_extra,
               header_data;

        msg_anum = number;
        msg_subject = new String("?");
        msg_author = new String("?");
        msg_date = new String("?");
        msg_id = new String("?");
        msg_refs = new String("?");
        msg_bytes = new String("?");
        msg_lines = new String("?");
        msg_extra = new String("?");

        try {
            headerFile = new FileReader( headerFileName );
            header = new BufferedReader( headerFile );
            do {
                headerLine = getLine( header );
                if ( headerLine != null ) {
                    htok = new StringTokenizer( headerLine );
                    try {
                        headerField = htok.nextToken();
                        if ( headerField.equalsIgnoreCase("Message-ID:")) {
                            msg_id = headerLine.substring( 12 );
                        }
                            // needs to handle multiple lines...
                        else if ( headerField.equalsIgnoreCase("References:")) {
                            msg_refs = headerLine.substring( 12 );
                        }
                        else if ( headerField.equalsIgnoreCase("Lines:")) {
                            msg_lines = headerLine.substring( 7 );
                        }
                        else if ( headerField.equalsIgnoreCase("Date:")) {
                            msg_date = headerLine.substring( 6 );
                        }
                        else if ( headerField.equalsIgnoreCase("From:")) {
                            msg_author = headerLine.substring( 6 );
                        }
                        else if ( headerField.equalsIgnoreCase("Subject:")) {
                            msg_subject = headerLine.substring( 8 );
                        }
                        else if ( headerField.equalsIgnoreCase("Xref:")) {
                            msg_extra = headerLine;
                        }
                    } catch( NoSuchElementException NSEe ) {}
                }
            } while ( headerLine != null );

//  <article number>\t<subject>\t<author>\t<date>\t<message-id>\t<references>\t<byte count>\t<line count>[\t<optional headers>]
                                // now save them.
            header_data = new String( msg_anum    + "\t" +
                                      msg_subject + "\t" +
                                      msg_author  + "\t" +
                                      msg_date    + "\t" +
                                      msg_id      + "\t" +
                                      msg_refs    + "\t" +
                                      msg_bytes   + "\t" +
                                      msg_lines            ); 

        } catch( FileNotFoundException FNFe ) {
            return ("File not found.");
        }
        return ( header_data );

    } // -----------------------------------------------------------------


    // -------------------------------------------------------------------
    private String getLine( BufferedReader input ) {
        String s;
        try {
            s = input.readLine();
        } catch( IOException any ) {
            s = null;
        }
        return s;
    } // -----------------------------------------------------------------


    // -------------------------------------------------------------------
    public String getLastResponse( ) {
        return ( server_response );
    } // -----------------------------------------------------------------


    // ------------------------------------------------------- textFollows
       /**
        *  Determine whether or not the last response from the server
        *  included textual data.  This is done by examining the
        *  response code at the start of the server's response string.
        *
        * @return A boolean value indicating whether or not additional
        *  lines of text are to follow immediately.
        ****/
    public boolean textFollows( ) {

        StringTokenizer response_fields = new StringTokenizer( server_response );
        int response_code = Integer.parseInt( response_fields.nextToken() );

        switch ( response_code ) {
            case NNTP_HELP_TEXT :  
            case NNTP_NEWSGROUP_LIST :
            case NNTP_XOVER_DATA_FOLLOWS :
            case NNTP_ARTICLE_FOLLOWS :
            case NNTP_HEADER_FOLLOWS :
            case NNTP_BODY_FOLLOWS :
            case NNTP_NEW_ARTICLES :
            case NNTP_NEW_NEWSGROUPS :
                return true;
            default : return false;
        }
    } // -----------------------------------------------------------------


    // ------------------------------------------------------- sendCommand
       /*
        * Sends the specified NNTP command to the news server. 
        *
        * Returns: The news server's response.
        */
    String sendCommand( String command ) {
        boolean end_of_data;
        String data_string;

        server_response = nntp_socket.transmit( command );

               // if text follows, open a temp. file and dump it in there.
        if ( textFollows() ) {
            temp_out = openTempFile();
            end_of_data = false;
            while (( ! end_of_data ) && ( ! stop_signalled )) {
                data_string = nntp_socket.readNextLine();
                if ( data_string.equals( NNTP_END_OF_TEXT ))
                    end_of_data = true;
                else {
                    if ( temp_out == null )
                        return ("IO Error.");
                    else {
                        temp_out.println( data_string );
                        socket_progress++;
                    }
                }
            }
            temp_out.close();
            socket_progress = 0;
            stop_signalled = false;
        }

        return ( server_response );
    } // -----------------------------------------------------------------


    // ------------------------------------------------------- getTempFile
       /**
        *  Obtain the name of the temporary file containing the most
        *  recent textual data sent by the server.  Whenever the server
        *  sends a response which contains multi-line text (such as
        *  a news article, a list of groups, etc.) the status message
        *  is returned as normal, but the following text is saved in a
        *  temporary text file.  Presumably, the news client will want
        *  to be able to retrieve this data, and thus will need the
        *  filename.  This method should be called <em>immediately</em>
        *  upon receiving a textual response, because NntpConnection
        *  won't preserve the filename.
        *
        * @return The filename of the temporary file.  Right now the
        *  name is always "nntp.tmp", but I may well change it in the
        *  future to assign dynamic filenames (in case I want to keep
        *  several of them around, e.g. for caching purposes, or for
        *  multi-threading).
        ****/
    public String getTempFile( ) {
        return ( temp_file );
    } // -----------------------------------------------------------------


    // ------------------------------------------------------ openTempFile
     /*
      * Creates a temporary file for dumping response text from the 
      * server into.  
      */
    private synchronized PrintWriter openTempFile( ) {
        temp_file = "nntp.tmp";
        try {
            FileOutputStream tfo = new FileOutputStream( temp_file );
            return ( new PrintWriter( tfo, true ));
        } catch( IOException IOe ) { return null; }
    } // -----------------------------------------------------------------


    // -------------------------------------------------------------------
       /**
        ****/
    public static boolean checkResponseCode( String response_string, 
                                             int desired_code        ) 
    {
        String code;
        boolean match;
        int value;

        match = false;
        StringTokenizer t = new StringTokenizer( response_string );
        try {
            code = t.nextToken();
        } catch( NoSuchElementException NSEe ) { code = response_string; }
        try {
            value = Integer.parseInt( code );
            if ( value == desired_code ) match = true;
                /*
                 * If a number can't be parsed, obviously there's no
                 * code-match - just do nothing and false will be returned.
                 *
                 */
        } catch( NumberFormatException NFe ) { ; }

        return ( match );        
    }


    public static boolean checkResponseCode( int actual, int desired ) {
        boolean match;

        match = false;
        if ( actual == desired ) match = true;

        return ( match );        
    } // -----------------------------------------------------------------


    // I really must define some custom exceptions, e.g. for
    //    socket timeout,
    //    broken/nonexistant socket,
    //    network unreachable, etc...

    // ======================================================== NntpSocket
    /**
     *  (Member class)
     *  
     *  A socket client which handles communications with an NNTP server.
     *  
     ****/
    class NntpSocket extends Object
    {
        Socket nntp_sock;
        InputStreamReader nntp_in;
        BufferedReader in;
        PrintWriter nntp_out;

        // --------------------------------------------------- constructor
        NntpSocket( ) {
            nntp_sock = null;
            nntp_in = null;
            nntp_out = null;
            in = null;
        } // -------------------------------------------------------------


        // ------------------------------------------------ openConnection
            /**
             *  Open a TCP connection to the specified host, on the
             *  specified port.
             *
             * @param nntp_host The name of the desired host.
             *
             * @param nntp_port The TCP port on which to connect.
             *
             * @return The (textual) response returned by the server, on
             *  connection.
             *  
             ****/
        public String openConnection( String nntp_host, int nntp_port ) 
                      throws IOException 
        {
            String response = null;

            System.out.println("Opening news socket...");
            try {
                nntp_sock = new Socket( nntp_host, nntp_port );

                    // Disable Nagle's algorithm so that progress monitors
                    // can work properly.  Doesn't really affect performance.
                nntp_sock.setTcpNoDelay( true );

                    // Set a read timeout value of 10 seconds.
                nntp_sock.setSoTimeout( 10000 );

                System.out.println("Connected to " +
                                   nntp_sock.getInetAddress() +
                                   " on port number " +
                                   nntp_sock.getPort() + "."    );

                nntp_in = new InputStreamReader( nntp_sock.getInputStream() );
                in = new BufferedReader( nntp_in );
                nntp_out = new PrintWriter( nntp_sock.getOutputStream(), true );

                response = getResponse();
            } catch( IOException any ) { System.err.println( any ); }

            return ( response );
        } // -------------------------------------------------------------


        // ------------------------------------------------------ transmit
        public synchronized String transmit( String command ) {
            String response = null;

            try {
//                System.out.println( command );
                nntp_out.println( command );
                response = getResponse();
            } catch( IOException any ) { System.err.println( any ); }

            return ( response );
        } // -------------------------------------------------------------


        // ---------------------------------------------------------------
        public synchronized String readNextLine( ) {
            String next_line;
            try {
                next_line = in.readLine();
            } catch( InterruptedIOException IIOe ) {
                next_line = ".";
            } catch( IOException IOe ) {
                next_line = ".";
            } catch( NullPointerException NPe ) {
                next_line =  ".";
            }
            return ( next_line );
        } // -------------------------------------------------------------


        // --------------------------------------------------- getResponse
        private synchronized String getResponse( )
        throws IOException
        // this needs to throw NullPointerException, too!!!
        {
            String response_string;

                       // get the status line (the first line of any response)
            response_string = readNextLine();
            if ( response_string.equals(".") )
                response_string = "599 Could not read socket (connection may have been lost)";

            return ( response_string );

        } // -------------------------------------------------------------


        // ----------------------------------------------- closeConnection
        public String closeConnection( ) {
            String response;

            response = transmit("QUIT");
            try { 
                in.close();
                nntp_in.close();
                nntp_out.close();
                nntp_sock.close();
            } catch( Exception any ) {
                System.err.println( any );
            }
            return ( response );
        } // -------------------------------------------------------------

    } // ======================[ End of NntpSocket ]======================

    
} // ======================[ End of NntpConnection ]======================
